"""
Asyncio
=======

This module contains various useful functions and classes for handling
common and awkward tasks in Asyncio.

"""

__all__ = [
    'gather_excs',
    'raise_first_exc',
    'to_async_iter',
    'to_sync_iter',
    'threadsafe_async_cache',
    'buffer_until_timeout',
    'BufferAsyncCalls',
    'ensure_aw',
    'loop_in_thread',
    'run_aw_threadsafe',
    'DaemonTask',
]

import sys
import queue
import logging
import asyncio as aio
from threading import Lock, Thread
from collections import defaultdict
from functools import partial, wraps
from concurrent.futures import ThreadPoolExecutor
from typing import (
    Iterator, Iterable, AsyncIterable, Callable, Set, Awaitable, Optional, Any,
    Type, Generic, Dict, DefaultDict, Union, TypeVar, FrozenSet, overload,
)

from .typing import T, Yields, AYields

E = TypeVar('E', bound=BaseException)
X = TypeVar('X')
Loop = aio.AbstractEventLoop

logger = logging.getLogger(__name__)

_DONE = object()


async def gather_excs(
    aws: Iterable[Awaitable[Any]],
    only: Type[E] = BaseException,  # type: ignore # MyPy bug
) -> AYields[E]:
    """
    Gather the given awaitables and yield any exceptions they raise.

    This is useful when spawning a lot of tasks and you only need to
    know if any of them error.

    >>> async def x():
    ...     await aio.sleep(0.01)

    >>> async def e():
    ...     await aio.sleep(0.01)
    ...     raise RuntimeError("This failed")

    >>> async def show_errors(excs: AYields[BaseException]):
    ...     async for exc in excs:
    ...         print(type(exc).__name__ + ':', *exc.args)

    >>> run = aio.get_event_loop().run_until_complete

    >>> run(show_errors(gather_excs([x(), e()])))
    RuntimeError: This failed

    A filter can be applied to the specific kind of exception to look
    for. Here, no ValueErrors were raised so there is no output:

    >>> run(show_errors(gather_excs([x(), e()], only=ValueError)))

    Switch it to RuntimeError or a parent class and it will show again:

    >>> run(show_errors(gather_excs([x(), e()], only=RuntimeError)))
    RuntimeError: This failed

    :param aws: Awaitables to gather
    :param only:
        Optional specific type of exceptions to filter on and yield
    """
    for res in await aio.gather(*aws, return_exceptions=True):
        if isinstance(res, only):
            yield res


async def raise_first_exc(aws: Iterable[Awaitable[Any]],
                          only: Type[BaseException] = BaseException) -> None:
    """
    Gather the given awaitables using :func:`gather_excs` and raise the
    first exception encountered.

    This is useful when you don't need the results of any tasks but need
    to know if there was at least one exception.

    >>> async def x():
    ...     await aio.sleep(0.01)

    >>> async def e():
    ...     await aio.sleep(0.01)
    ...     raise RuntimeError("This failed")

    >>> run = aio.get_event_loop().run_until_complete

    >>> run(raise_first_exc([x(), e()]))
    Traceback (most recent call last):
    ...
    RuntimeError: This failed

    A filter can be applied to the specific kind of exception to look
    for. Here, no ValueErrors were raised so there is no output:

    >>> run(raise_first_exc([x(), e()], only=ValueError))

    Switch it to RuntimeError or a parent class and it will raise again:

    >>> run(raise_first_exc([x(), e()], only=RuntimeError))
    Traceback (most recent call last):
    ...
    RuntimeError: This failed

    :param aws: Awaitables to gather
    :param only:
        Optional specific type of exceptions to filter on and raise
    """
    async for exc in gather_excs(aws, only):
        raise exc


async def to_async_iter(iterable: Iterable[T]) -> AYields[T]:
    """
    Convert the given iterable from synchronous to asynchrounous by
    iterating it in a background thread.

    This can be useful for interacting with libraries for APIs that do
    not have an asynchronous version yet.

    To demonstrate this, let's create a function using Requests which
    makes a synchronous calls to the free {JSON} Placeholder API and
    then yields the results:

    >>> import requests

    >>> def user_names() -> Yields[str]:
    ...     for uid in range(3):
    ...         url = f'https://jsonplaceholder.typicode.com/users/{uid + 1}'
    ...         yield requests.get(url).json()['name']

    Now, we can convert this to an async iterator which won't block the
    event loop:

    >>> async def print_user_names():
    ...     async for name in to_async_iter(user_names()):
    ...         print(name)

    >>> aio.get_event_loop().run_until_complete(print_user_names())
    Leanne Graham
    Ervin Howell
    Clementine Bauch
    """

    # Quick optimization: if it isn't an iterator, just yield from
    # it directly to avoid using a background thread
    if not isinstance(iterable, Iterator):
        for x in iterable:
            yield x
        return

    def _queue_elements() -> None:
        try:
            for x in iterable:
                put(x)
        finally:
            put(_DONE)

    q: 'aio.Queue[Union[T, object]]' = aio.Queue()
    loop = aio.get_event_loop()
    put = partial(loop.call_soon_threadsafe, q.put_nowait)
    with ThreadPoolExecutor(1) as pool:
        future = loop.run_in_executor(pool, _queue_elements)
        while True:
            i = await q.get()
            if i is _DONE:
                break
            yield i  # type: ignore
        await future  # Bubble any errors


def to_sync_iter(iterable: AsyncIterable[T],
                 *,
                 loop: Optional[Loop] = None) -> Yields[T]:
    """
    Convert the given iterable from asynchronous to synchrounous by
    by using a background thread running a new event loop to iterate it.

    This can be useful for providing a synchronous interface to a new
    async library for backwards compatibility.

    To demonstrate this, let's create a function using HTTPX which
    makes asynchronous calls to the free {JSON} Placeholder API and
    then yields the results:

    >>> import httpx

    >>> async def user_name(uid: int) -> str:
    ...     url = f'https://jsonplaceholder.typicode.com/users/{uid + 1}'
    ...     async with httpx.AsyncClient() as client:
    ...         resp = await client.get(url)
    ...         return resp.json()['name']

    >>> async def user_names() -> AYields[str]:
    ...     for name in await aio.gather(*map(user_name, range(3))):
    ...         yield name

    Now, we can convert this to a synchronous iterator and show the
    results:

    >>> for name in to_sync_iter(user_names()):
    ...     print(name)
    Leanne Graham
    Ervin Howell
    Clementine Bauch

    :param iterable: Asynchonrous iterable to process
    :param loop: Optional specific loop to use to process the iterable
    """
    if loop is None:
        loop = aio.new_event_loop()

    def _set_loop_and_queue_elements(_loop: Loop) -> None:
        aio.set_event_loop(_loop)
        _loop.run_until_complete(_queue_elements())

    async def _queue_elements() -> None:
        try:
            async for x in iterable:
                put(x)
        finally:
            put(_DONE)  # type: ignore

    q: 'queue.Queue[T]' = queue.Queue()
    put = q.put_nowait
    with ThreadPoolExecutor(1) as pool:
        future = pool.submit(_set_loop_and_queue_elements, loop)
        try:
            while True:
                i = q.get()
                if i is _DONE:
                    break
                yield i
        finally:
            future.result()


def threadsafe_async_cache(
    func: Callable[..., Awaitable[T]],
) -> Callable[..., Awaitable[T]]:
    """
    Simple thread-safe asynchronous cache decorator which ensures that
    the decorated/wrapped function is only ever called once for a given
    set of inputs even across different event loops in multiple threads.

    To demonstrate this, we'll define an async function which sleeps
    to simulate work and then returns the square of the input:

    >>> import asyncio as aio

    >>> @threadsafe_async_cache
    ... async def square(x: int):
    ...     await aio.sleep(0.1)
    ...     print("Squaring:", x)
    ...     return x * x

    Then we'll define a function that will use a new event loop to call
    the function 10 times in parallel:

    >>> def call_func(x: int):
    ...     loop = aio.new_event_loop()
    ...     aio.set_event_loop(loop)
    ...     return loop.run_until_complete(aio.gather(*(
    ...         square(x) for _ in range(10)
    ...     )))

    And finally we'll call that function 10 times in parallel across 10
    threads:

    >>> with ThreadPoolExecutor(max_workers=10) as pool:
    ...     results = list(pool.map(call_func, [2] * 10))
    Squaring: 2

    As can be seen above, the function was only called once despite
    being invoked 100 times across 10 threads and 10 event loops. This
    can be confirmed by counting the total number of results:

    >>> len([x for row in results for x in row])
    100

    .. warning::
        This cache has no max size and can very easily be the source of
        memory leaks. I typically use this to wrap object methods on a
        per-instance basis during intilaztion so that the cache only
        lives as long as the object.
    """
    cache: Dict[FrozenSet[Any], T] = {}
    # 1 lock per input key
    locks: DefaultDict[FrozenSet[Any], Lock] = defaultdict(Lock)

    @wraps(func)
    async def _wrapper(*args: Any, **kwargs: Any) -> T:
        # Get the key from the input arguments
        key = frozenset(args + tuple(kwargs.values()))
        while True:
            try:  # to get the value from the cache
                return cache[key]
            except KeyError:
                pass
            # Need to calculate and cache the value once
            lock = locks[key]
            if lock.acquire(blocking=False):  # First to arrive, cache
                try:  # to run the function and get the result
                    result = await func(*args, **kwargs)
                except BaseException:
                    raise  # Don't cache errors, maybe timeouts or such
                else:  # Successfully got the result
                    cache[key] = result
                    del locks[key]  # Allow lock to be garbage collected
                finally:  # Ensure lock is always released
                    lock.release()
                return result
            # Wait for the other thread/task to cache the result by
            # acquiring the lock in another thread to avoid blocking the
            # current loop. Once this is done, the loop will continue
            # and the cached result can be retrieved.
            await aio.get_event_loop().run_in_executor(None, lock.acquire)
            lock.release()

    return _wrapper


_BufferFunc = Callable[[Set[T]], Awaitable[None]]


@overload
def buffer_until_timeout(
    *,
    timeout: float = 1,
) -> 'Callable[[_BufferFunc[T]], BufferAsyncCalls[T]]':
    ...


@overload
def buffer_until_timeout(
    func: _BufferFunc[T],
    *,
    timeout: float = 1,
) -> 'BufferAsyncCalls[T]':
    ...


def buffer_until_timeout(
    func: Optional[_BufferFunc[T]] = None,
    *,
    timeout: float = 1,
) -> Union['BufferAsyncCalls[T]',
           'Callable[[_BufferFunc[T]], BufferAsyncCalls[T]]']:
    """
    Async function decorator/wrapper which buffers the arg passed in
    each call. After a given timeout has passed since the last call,
    the function is invoked on the event loop that the function was
    decorated nearest.

    >>> @buffer_until_timeout(timeout=0.1)
    ... async def buffer(args: Set[int]):
    ...     print("Buffered:", args)

    >>> for i in range(5):
    ...     buffer(i)

    >>> aio.get_event_loop().run_until_complete(aio.sleep(0.5))
    Buffered: {0, 1, 2, 3, 4}

    Another function can also wait for all elements to be processed
    before continueing:

    >>> async def after_buffered():
    ...     await buffer.wait()
    ...     print("All buffered!")

    >>> for i in range(5):
    ...     buffer(i)

    >>> aio.get_event_loop().run_until_complete(after_buffered())
    Buffered: {0, 1, 2, 3, 4}
    All buffered!

    :param func: Function to wrap
    :param timeout:
        Number of seconds to wait after the most recent call before
        executing the function.
    :return: Non-async function which will buffer the given argument
    """
    if func is None:
        return partial(buffer_until_timeout, timeout=timeout)  # type: ignore
    return wraps(func)(BufferAsyncCalls(func, timeout=timeout))


class BufferAsyncCalls(Generic[T]):
    """
    Wrapper around an async callable to buffer inputs before calling
    it after a given timeout. This class should not be initialized
    directly, users should use :func:`buffer_until_timeout` instead.

    .. warning::
        While this object is thread-safe and arguments can be added from
        anywhere, the loop this object was created in must be running
        for it to work.
    """

    def __init__(self, func: _BufferFunc[T], *, timeout: float = 1):
        #: Wrapped function which should take a set and will be called
        #: once inputs are buffered and timeout is reached
        self.func = func
        #: Timeout in seconds to wait after the last element from the
        #: queue before calling the function
        self.timeout = timeout

        #: Event loop that will be used to get args from the queue and
        #: execute the function
        self.loop = aio.get_event_loop()

        #: Asyncio queue which new arguments will be placed in
        #: Arguments are passed as async iterables (the most complex
        #: way of providing arguments) to ensure all elements are
        #: processed on time while avoiding deadlocks.
        self.q: 'aio.Queue[AsyncIterable[T]]' = aio.Queue()
        #: Asyncio event that is set when all current arguments have
        #: been processed and cleared whenever a new one is added.
        self.event = aio.Event()
        self.event.set()

        #: Task which is infinitely processing the queue
        self._waiting = DaemonTask(
            self._waiter(),
            loop=self.loop,
            name=f"Buffering {self.func!r}",
        )
        #: Current task that is waiting for a new element from the queue
        self._getting: Optional['aio.Task[AsyncIterable[T]]'] = None

    def __call__(self, _arg: T) -> None:
        """
        Place the given argument on the queue to be processed on the
        next execution of the function.
        """
        self._put(_obj_to_aiter(_arg))

    def await_(self, _arg: Awaitable[T]) -> None:
        """
        Schedule the given awaitable to be be put onto the queue after
        it has been awaited.

        >>> @buffer_until_timeout(timeout=0.1)
        ... async def buffer(args: Set[int]):
        ...     print("Buffered:", args)

        >>> async def delay(x: T) -> T:
        ...     await aio.sleep(0)
        ...     return x

        >>> for i in range(5):
        ...     buffer.await_(delay(i))

        >>> aio.get_event_loop().run_until_complete(buffer.wait())
        Buffered: {0, 1, 2, 3, 4}
        """
        self._put(_awaitable_to_aiter(_arg))

    def map(self, _args: Iterable[T]) -> None:
        """
        Place an iterable of args onto the queue to be processed.

        >>> @buffer_until_timeout(timeout=0.1)
        ... async def buffer(args: Set[int]):
        ...     print("Buffered:", args)

        >>> buffer.map(range(5))

        >>> aio.get_event_loop().run_until_complete(buffer.wait())
        Buffered: {0, 1, 2, 3, 4}

        """
        self._put(to_async_iter(_args))

    def amap(self, _args: AsyncIterable[T]) -> None:
        """
        Schedule an async iterable of args to be put onto the queue.

        >>> @buffer_until_timeout(timeout=0.1)
        ... async def buffer(args: Set[int]):
        ...     print("Buffered:", args)

        >>> buffer.amap(to_async_iter(range(5)))

        >>> aio.get_event_loop().run_until_complete(buffer.wait())
        Buffered: {0, 1, 2, 3, 4}

        """
        self._put(_args)

    async def wait(self, *, cancel: bool = True) -> None:
        """
        Wait for the event to be set indicating that all arguments
        currently in the queue have been processed.

        :param cancel:
            If True (the default) and there is a task currently waiting
            for a new item from the queue, cancel it so that the queued
            arguments are processed immediately instead of after the
            timeout. Settings this to False will force the full timeout
            to be met before the function is called if it hasn't been
            already.
        """
        # Process all queued tasks
        # Create a task because some unknown bug on 3.7 blocks
        # infinitely sometimes if we await join directly
        await self.loop.create_task(self.q.join())
        if cancel and self._getting and not self._getting.done():
            # Cancel the current q.get to avoid waiting for timeout
            # The sleep(0) forces task switching to ensure that
            # _process_queue gets at least one cycle to pull remaining
            # elements off the queue
            await aio.sleep(0)
            self._getting.cancel()
        # Wait for the function to finish processing
        await self.event.wait()

    async def wait_from_anywhere(self, *, cancel: bool = True) -> None:
        """
        Wrapper around :meth:`wait` which uses :func:`ensure_aw` to
        handle waiting from a possibly different event loop.
        """
        return await ensure_aw(self.wait(cancel=cancel), self.loop)

    async def _waiter(self) -> None:
        """
        Simple loop which tries to process the queue infinitely using
        :meth:`_process_queue`. This is spawned automatically upon
        """
        while True:
            await self._process_queue()

    async def _process_queue(self) -> None:
        """
        Internal method to retrieve all current elements from queue and
        execute the function on timeout or cancellation.
        """
        inputs: Set[T] = set()

        async def _load_inputs(iterable: AsyncIterable[T]) -> None:
            try:
                async for i in iterable:
                    inputs.add(i)
            except BaseException:  # noqa
                logger.exception("Failed to get args from: %r", iterable)

        # Get first element, block infinitely until one appears
        try:
            input_gens = [_load_inputs(await self.q.get())]
        except RuntimeError:  # Most likely loop shutdown
            return
        else:
            self.event.clear()  # Ensure cleared in case previous cancel
            self.q.task_done()
        # Keep adding new args until the function has run successfully
        while not self.event.is_set():
            # Get all the input generators currently in the queue
            input_gens.extend(map(_load_inputs, self._empty_queue()))
            # Schedule the q.get() and save it as an attribute so it
            # can be cancelled as necessary. This needs to be scheduled
            # *before* waiting for the known inputs.
            self._getting = self._schedule_with_timeout(self.q.get())
            if input_gens:  # Load as many as possible concurrently
                await aio.gather(*input_gens)
                input_gens.clear()  # Clear processed input generators
            # Now wait for the next arguments to buffer. If this times
            # out or is cancelled, its time to run the function.
            try:
                await _load_inputs(await self._getting)
            except (aio.TimeoutError, aio.CancelledError):
                await self._run_func(inputs)
            else:
                self.q.task_done()

    async def _run_func(self, inputs: Set[T]) -> None:
        """
        Run :attr:`func` with the given set of inputs and set
        :attr:`event` once it has finished successfully.

        If an exception is raised, log it with its traceback and return
        without setting the event to prevent the buffered inputs from
        being lost.
        """
        try:
            if inputs:  # Could be empty if all empty iterators
                await self.func(inputs)
        except BaseException as e:  # noqa
            logging.exception("Failed to run %s, retrying", self.func)
        else:
            self.event.set()

    def _schedule_with_timeout(self, coro: Awaitable[X]) -> 'aio.Task[X]':
        """
        Helper method to create a task for the given coroutine with the
        configured :attr:`timeout`.
        """
        return self.loop.create_task(aio.wait_for(coro, self.timeout))

    def _put(self, iterable: AsyncIterable[T]) -> None:
        """
        Helper method to put an async iterable onto the queue and clear
        the event.
        """
        self.event.clear()
        self.loop.call_soon_threadsafe(self.q.put_nowait, iterable)

    def _empty_queue(self) -> Yields[AsyncIterable[T]]:
        """Get all of the elements immediately available in the queue."""
        while True:
            try:
                yield self.q.get_nowait()
            except aio.QueueEmpty:
                break
            else:
                self.q.task_done()


_CROSS_LOOP_POOL = ThreadPoolExecutor(32)


async def ensure_aw(aw: Awaitable[T], loop: Loop) -> T:
    """
    Return an awaitable for the running loop which will ensure the
    given awaitable is evaluated in the given loop. This is useful for
    waiting on events or tasks defined in another thread's event loop.

    .. note::
        The situation in which this function is necessary should be
        avoided when possible, but sometimes it is not always possible.

    >>> e = aio.Event()
    >>> e.set()

    >>> e_loop = aio.get_event_loop()
    >>> new_loop = aio.new_event_loop()
    >>> e_loop is not new_loop
    True

    >>> def task(): return e_loop.create_task(e.wait())

    >>> run = new_loop.run_until_complete
    >>> run(task())  # ValueError: Belongs to a different loop
    Traceback (most recent call last):
    ...
    ValueError: ...

    If the target loop isn't running, it will be run directly using
    another thread:

    >>> run(ensure_aw(task(), e_loop))
    True

    If the target loop is running, it will be used as is using
    :func:`run_aw_threadsafe` internally:

    >>> stop = loop_in_thread(e_loop)

    >>> run(ensure_aw(task(), e_loop))
    True

    >>> stop()  # Cleanup the loop
    """

    main_loop = aio.get_event_loop()

    if main_loop is loop:
        return await aw

    if loop.is_running():
        return await run_aw_threadsafe(aw, loop)

    if loop.is_closed():
        raise RuntimeError("Target loop is closed!")

    def _loop_thread() -> T:
        aio.set_event_loop(loop)
        return loop.run_until_complete(aw)

    return await main_loop.run_in_executor(_CROSS_LOOP_POOL, _loop_thread)


def loop_in_thread(loop: Loop) -> Callable[[], None]:
    """
    Start the given loop in a thread, let it run indefinitely, and
    return a function which can be called to signal the loop to stop.

    >>> from time import sleep

    >>> loop = aio.get_event_loop()

    >>> stop = loop_in_thread(loop)
    >>> sleep(0.05)  # Give the loop a chance to start
    >>> loop.is_running()  # Running in background thread
    True

    >>> loop.call_soon_threadsafe(print, "called")  # doctest: +SKIP
    <Handle print('called')>

    .. testsetup::

        Run after delay to improve test case timing consistency

        >>> loop.call_soon_threadsafe(lambda: sleep(0.01) or print('called'))
        <Handle ...lambda...>

    >>> sleep(0.05)  # Give the loop a chance to process
    called

    >>> stop()  # Signal loop to stop

    >>> sleep(0.05)  # Give it a little bit of time to actually stop

    >>> loop.is_running()  # No longer running
    False
    """
    stop = aio.Event()

    async def _spin() -> None:
        aio.set_event_loop(loop)
        await stop.wait()

    Thread(target=loop.run_until_complete, args=(_spin(),), daemon=True).start()

    def _stopper() -> None: loop.call_soon_threadsafe(stop.set)

    return _stopper


async def run_aw_threadsafe(aw: Awaitable[T], loop: Loop) -> T:
    """
    Thin wrapper around :func:`asyncio.run_coroutine_threadsafe` to
    handle any awaitable.

    .. warning::
        This does not extra handling of any event loop conflicts. Use
        :func:`ensure_aw` for that.
    """
    fut = aio.run_coroutine_threadsafe(_aw_to_coro(aw), loop)
    return await aio.wrap_future(fut)


async def _aw_to_coro(aw: Awaitable[T]) -> T:
    """Wrap a given awaitable so it appears as a coroutine."""
    return await aw


async def _obj_to_aiter(o: T) -> AsyncIterable[T]:
    """
    Wrap a given sync object so it is a single element, async iterable
    for compatibility with :attr:`BufferAsyncCalls.q`.
    """
    yield o


async def _awaitable_to_aiter(o: Awaitable[T]) -> AsyncIterable[T]:
    """
    Wrap a awaitable object so it is a single element, async iterable
    for compatibility with :attr:`BufferAsyncCalls.q`.
    """
    yield await o


class DaemonTask(aio.Task):  # type: ignore
    """
    Custom :class:`asyncio.Task` which is meant to run forever and
    therefore doesn't warn when it is still pending at loop shutdown.
    """

    if sys.version_info < (3, 8):

        # Ignore name arg when it isn't available for compatibility
        def __init__(self, coro, *, loop=None, name=None):
            super().__init__(coro, loop=loop)

    # Skip the __del__ defined by aio.Task which does the logging and
    # then calls super()
    __del__ = aio.Task.__base__.__del__  # type: ignore



"""
Itertools
=========

This module contains various useful functions for working with iterators
beyond the scope the itertools library without overlapping with
more_itertools.

"""

__all__ = [
    'exhaust',
    'split',
]

import operator as op
from collections import deque
from itertools import tee, compress
from typing import Tuple, Iterable, Iterator, Callable, Union, Any

from .typing import T


def exhaust(iterable: Iterable[Any]) -> None:
    """
    Iterate through all the values of an iterable. This is useful for
    applying a mapped function which doesn't return a useful value
    without writing a dedicated for loop.

    >>> exhaust(map(print, range(3)))
    0
    1
    2

    """
    # deque is ~10% faster than writing a loop
    deque(iterable, maxlen=0)


def split(
    iterable: Iterable[T],
    condition: Union[Iterable[bool], Callable[[T], bool]],
) -> Tuple[Iterator[T], Iterator[T]]:
    """
    Similar to :func:`itertools.compress` except that it returns two
    iterators: one where with the values where the condition is True
    and one where with the values where the condition is False.

    >>> from itertools import cycle
    >>> evens, odds = split(range(5), cycle((True, False)))
    >>> list(evens)
    [0, 2, 4]
    >>> list(odds)
    [1, 3]

    The condition can also be a boolean callable:

    >>> evens, odds = split(range(5), lambda x: x % 2 == 0)
    >>> list(evens)
    [0, 2, 4]
    >>> list(odds)
    [1, 3]

    :param iterable:
        Iterable to split into two sub-iterators.
        If given as iterator, it must not be advanced directly later!
    :param condition:
        Either a boolean callable to map to each element or a boolean
        iterable indicating which values belong in which set.
        If given as iterator, it must not be advanced directly later!
    :return: Tuple of two iterators: (where_true, where_false)
    """
    if callable(condition):
        iterable, ci = tee(iterable)
        condition = map(condition, ci)
    i1, i2 = tee(iterable)
    c1, c2 = tee(condition)
    return compress(i1, c1), compress(i2, map(op.not_, c2))



"""
Typing
======

This module contains various useful typing definitions to simplify
using some of the advanced typing features and avoid creating the same
generic declarations from project to project.

It also fallbacks for the following common new typing features to avoid
import errors on older versions of Python:

    - ``Literal``
    - ``Protocol``
    - ``TypedDict``

"""

__all__ = [
    'Literal', 'Protocol', 'TypedDict',
    'MaybeIter', 'MaybeAwaitable', 'Yields', 'AYields',
    'T', 'T_co', 'T_contra', 'KT', 'VT', 'KT_co', 'VT_co',
]

import logging
from typing import (
    TypeVar, Dict, Union, Iterable, Awaitable,
    Generator, AsyncGenerator,
)

logger = logging.getLogger(__name__)


try:
    from typing import Literal, Protocol, TypedDict
except ImportError:
    try:
        from typing_extensions import (  # type: ignore
            Literal, Protocol, TypedDict,
        )
    except ImportError:
        # Shim implement common back ports to avoid requiring
        # typing_extensions to be installed for regular execution.

        # Note that this is not a replacement for typing_extensions which
        # should be made a normal dependency of any packages that use newer
        # typing features and wish to support older Python versions.
        logger.info("Creating shims for common typing features."
                    " Please install typing_extensions to avoid this.")
        Literal = type('Literal', (type,),  # type: ignore
                       {'__getitem__': lambda c, i: c})
        Protocol = type('Protocol', (type,),   # type: ignore
                        {'__getitem__': lambda c, i: c})
        TypedDict = type('TypedDict', (Dict,), {})

# Define common TypeVar instances for convenience

#: Generic invariant type var
T = TypeVar('T')
#: Generic covariant type var
T_co = TypeVar('T_co', covariant=True)
#: Generic contravariant type var
T_contra = TypeVar('T_contra', contravariant=True)

#: Generic invariant type var for keys
KT = TypeVar('KT')
#: Generic invariant type var for values
VT = TypeVar('VT')

#: Generic covariant type var for keys
KT_co = TypeVar('KT_co', covariant=True)
#: Generic covariant type var for values
VT_co = TypeVar('VT_co', covariant=True)

#: Generic type for an object which may be a single value of a given
#: type or an iterable of that type
MaybeIter = Union[T, Iterable[T]]

#: Generic type for an object which could be wrapped in an awaitable
MaybeAwaitable = Union[T, Awaitable[T]]

#: Generic type for a Generator which only yields
#:
#: This is different from ``Iterator[T]`` as it also declares that the
#: object implements the rest ``Generator`` interface such as
#: ``gen.close()`` and ``gen.throw()``.
Yields = Generator[T, None, None]

#: Generic type for an ``AsyncGenerator`` which only yields
AYields = AsyncGenerator[T, None]



"""
FileLock
========

Thread-safe, platform independent file lock supporting the interface
as the standard :class:`threading.Lock`.

This code is loosely adapted from the `py-filelock
<https://github.com/benediktschmitt/py-filelock/blob
/b30bdc4fb998f5f4350257235eb51147f9e81862/filelock.py>`__
module, which is unfortunately not threadsafe.

.. warning::
    Unlike py-filelock, this module does not provide a soft file lock
    to avoid any misinterpretation of the safety of using it.

Example Usage:

.. code-block:: python

    from aiuti.filelock import FileLock

    flock = FileLock('/path/to/file.lock')

    with flock:
        # Use some resource with exclusive thread and process access

"""

__all__ = [
    'BaseFileLock',
    'WindowsFileLock',
    'UnixFileLock',
    'FileLock',
]

import os
import abc
import time
import logging
import threading
import contextlib
from typing import Union, Optional, TypeVar, ClassVar, Any

from .typing import Yields


def _import_optional(name: str) -> Any:
    try:
        return __import__(name)
    except ImportError:
        pass


msvcrt = _import_optional('msvcrt')
fcntl = _import_optional('fcntl')

_logger = logging.getLogger(__name__)

PathLike = Union[str, os.PathLike]
FileLockT = TypeVar('FileLockT', bound='BaseFileLock')


class BaseFileLock(abc.ABC):
    """Abstract base class for a file lock."""

    _thread_lock: Union[threading.Lock, threading.RLock]

    def __init__(self,
                 lock_file: PathLike,
                 timeout: float = -1,
                 reentrant: bool = False):
        """
        :param lock_file: Path to the lock file
        :param timeout:
            Optional timeout, in seconds, after which a :class:`Timeout`
            exception will be raised indicating the lock couldn't be
            acquired.
        :param reentrant:
            If True, use an :class:`~threading.RLock` internally to
            allow the same thread to acquire the lock multiple times.
        """
        # The path to the lock file.
        self._lock_file: PathLike = lock_file

        # The file descriptor for the *_lock_file* as it is returned by
        # os.open(). This is only NOT None if the lock is held.
        self._lock_file_fd: Optional[int] = None

        # The default timeout value.
        self.timeout: float = timeout

        # Boolean indicating if the lock is recursive, thread-wise
        self._reentrant: bool = reentrant

        # The internal threading lock to hold while waiting for and
        # while the file lock is held
        if self._reentrant:
            self._thread_lock = threading.RLock()
        else:
            self._thread_lock = threading.Lock()

        # For reentrant locks, the number of levels deep. When this
        # falls to zero, the file lock can be released
        self._lock_counter: int = 0

    @property
    def lock_file(self) -> PathLike:
        """The path to the lock file."""
        return self._lock_file

    @property
    def is_locked(self) -> bool:
        """True, if the object holds the file lock."""
        return self._lock_file_fd is not None

    def acquire(self,
                blocking: bool = True,
                timeout: Optional[float] = None,
                poll_interval: float = 0.05) -> bool:
        """
        Acquire the file lock.

        :param blocking:
            If True, block until the file lock is acquired or the
            timeout is exceeded. If False, return immediately if the
            lock cannot be acquired.
        :param timeout:
            Optional override for :attr:`timeout`, the maximum number of
            seconds to wait for the file lock.
        :param poll_interval:
            Number of seconds between attempts to acquire the file lock.
        :return: True if lock acquired, False otherwise
        """
        # Use the default timeout, if no timeout is provided.
        if timeout is None:
            timeout = self.timeout if blocking else -1
        else:
            blocking = blocking if timeout < 0 else True

        lid = id(self)
        fn = self._lock_file

        if not self._thread_lock.acquire(blocking, timeout):
            _logger.debug('Timeout on acquiring thread lock %s on %s', lid, fn)
            return False

        self._lock_counter += 1  # Keep lock counter synced with RLock

        if self.is_locked:
            return True

        start_time = time.time()

        def _cleanup_thread_lock() -> None:
            self._decrement_lock_counter()
            self._thread_lock.release()

        try:
            while True:
                _logger.debug('Attempting to acquire lock %s on %s', lid, fn)
                self._acquire(block=blocking and timeout < 0)

                if self.is_locked:
                    _logger.info('Lock %s acquired on %s', lid, fn)
                    break
                elif not blocking:
                    _logger.debug('Failed to acquire lock %s on %s', lid, fn)
                    _cleanup_thread_lock()
                    return False
                elif 0 <= timeout < time.time() - start_time:
                    _logger.debug('Timeout on acquiring lock %s on %s', lid, fn)
                    _cleanup_thread_lock()
                    return False
                else:
                    _logger.debug(
                        'Lock %s not acquired on %s, waiting %s seconds ...',
                        lid, fn, poll_interval,
                    )
                    time.sleep(poll_interval)
        except:  # noqa
            _logger.exception("Failed to acquire lock %s on %s", lid, fn)
            _cleanup_thread_lock()
            raise

        return True

    @contextlib.contextmanager
    def acquire_ctx(self,
                    blocking: bool = True,
                    timeout: Optional[float] = None,
                    poll_interval: float = 0.05) -> Yields[None]:
        """
        Context manager to make it more convenient to pass custom args
        to :meth:`acquire`.

        :raises TimeoutError: If :meth:`acquire` returns ``False``.
        """
        if not self.acquire(blocking, timeout, poll_interval):
            raise TimeoutError("Failed to acquire file lock:", self._lock_file)
        try:
            yield
        except BaseException:  # noqa
            self.release()

    def release(self, force: bool = False) -> None:
        """
        Release the file lock.

        .. note::
            The file lock is only completely released if the lock
            counter is zero when using reentrant locks.

        .. note::
            The lock file itself is not automatically deleted to avoid
            race conditions.

        :param force:
            If true, the lock counter is ignored and the lock is
            released in every case.
        """
        if not self.is_locked:
            return

        self._decrement_lock_counter()

        if self._lock_counter == 0 or force:
            lid = id(self)
            fn = self._lock_file

            _logger.debug('Attempting to release lock %s on %s', lid, fn)
            try:
                self._release()
            except:  # noqa
                _logger.exception("Failed to release lock %s on %s", lid, fn)
            else:
                self._lock_counter = 0
                _logger.info('Lock %s released on %s', lid, fn)

        try:
            self._thread_lock.release()
        except RuntimeError:  # not reentrant and already unlocked
            pass

    # Open mode for the file descriptor
    _FD_OPEN_MODE: ClassVar[int] = os.O_RDWR | os.O_CREAT | os.O_TRUNC

    def _acquire(self, block: bool = True) -> None:
        """
        Attempt to acquire the platform dependent lock and set
        :attr:`_lock_file_fd` to the file descriptor of the lock file.

        :param block:
            If True, attempt to block until the lock is acquired.
        """
        try:
            fd = os.open(self._lock_file, self._FD_OPEN_MODE)
        except OSError:
            return
        try:
            self._lock(fd, block)
        except (IOError, OSError):
            os.close(fd)
        else:
            self._lock_file_fd = fd

    def _release(self) -> None:
        """
        Release the platform dependent lock and clear
        :attr:`_lock_file_fd`
        """
        fd, self._lock_file_fd = self._lock_file_fd, None
        assert isinstance(fd, int)
        try:
            self._unlock(fd)
        finally:
            os.close(fd)

    @abc.abstractmethod
    def _lock(self, fd: int, block: bool = True) -> None:
        """
        Lock the platform dependent lock for the file descriptor.

        :param fd: File descriptor to acquire a file lock for
        :param block:
            If True, attempt to block until the lock is acquired.
        """

    @abc.abstractmethod
    def _unlock(self, fd: int) -> None:
        """
        Unlock the platform dependent lock for the file descriptor.

        .. note::
            This method should not remove the lock file to avoid race
            conditions.
        """

    def _decrement_lock_counter(self) -> None:
        """
        Helper method to decrement the lock counter without letting it
        drop below 0.
        """
        self._lock_counter = max(0, self._lock_counter - 1)

    def __enter__(self: FileLockT) -> FileLockT:
        self.acquire()
        return self

    def __exit__(self, *_exc: Any) -> None:
        self.release()

    def __del__(self) -> None:
        self.release(force=True)


class WindowsFileLock(BaseFileLock):
    """
    Windows specific file lock implementation which uses
    :func:`msvcrt.locking` to hard lock the lock file.
    """

    def _lock(self, fd: int, block: bool = True) -> None:
        msvcrt.locking(fd, msvcrt.LK_LOCK if block else msvcrt.LK_NBLCK, 1)

    def _unlock(self, fd: int) -> None:
        msvcrt.locking(fd, msvcrt.LK_UNLCK, 1)


class UnixFileLock(BaseFileLock):
    """
    Unix specific file lock implementation which uses
    :func:`fcntl.flock` to hard lock the lock file.
    """

    def _lock(self, fd: int, block: bool = True) -> None:
        fcntl.flock(fd,  fcntl.LOCK_EX | (0 if block else fcntl.LOCK_NB))

    def _unlock(self, fd: int) -> None:
        fcntl.flock(fd, fcntl.LOCK_UN)


class _UnsupportedFileLock(BaseFileLock):

    def _lock(self, fd: int, block: bool = True) -> None:
        raise RuntimeError("Filelock isn't supported by your platform!")

    def _unlock(self, fd: int) -> None:
        raise RuntimeError("Filelock isn't supported by your platform!")


#: Alias for the lock, which should be used for the current platform.
#: On Windows, this is an alias for :class:`WindowsFileLock` and on Unix
#: for :class:`UnixFileLock`.
FileLock = _UnsupportedFileLock

if msvcrt:
    FileLock = WindowsFileLock  # type: ignore
elif fcntl:
    FileLock = UnixFileLock  # type: ignore
else:
    logging.warning("No platform specific lock available!")



"""
Parsing
=======

This module contains various useful tools for parsing input data.

"""

__all__ = ['parse_to_dict']

import ast
from typing import Iterable, Union, Callable, Any, Tuple, Mapping, Dict


def parse_to_dict(items: Union[Mapping[Any, Any],
                               Iterable[Union[Tuple[Any, ...], str]]],
                  *,
                  sep: str = '=',
                  parse: Callable[[str], Any] = ast.literal_eval,
                  parse_keys: bool = True) -> Dict[Any, Any]:
    """
    Normalize the given input into a dictionary and parse keys & values
    as necessary into literal objects.

    When given an iterable of strings, they are assumed to be in the
    form of ``<key><sep><value>`` like ``a=b`` and will be parsed as
    such.

    Keys or values which fail to parse will be retained as is.

    By default, both keys and values are parsed literally:

    >>> parse_to_dict({'a': '1', '2': '"b"'})
    {'a': 1, 2: 'b'}

    >>> parse_to_dict(['a=1', '2="b"', '"b"=1.4'])
    {'a': 1, 2: 'b', 'b': 1.4}

    Parsing keys can be disabled with the ``parse_keys`` parameter:

    >>> parse_to_dict({'a': '1', '2': '"b"'}, parse_keys=False)
    {'a': 1, '2': 'b'}

    >>> parse_to_dict(['a=1', '2=b'], parse_keys=False)
    {'a': 1, '2': 'b'}

    The default separator can be changed using the ``sep`` parameter:

    >>> parse_to_dict(['a:1', '2:"b"', '"b":1.4'], sep=':')
    {'a': 1, 2: 'b', 'b': 1.4}

    :param items: Mapping or iterable to parse to a dictionary
    :param sep:
        For elements that are strings, the separator between the key and
        value
    :param parse:
        Function which will be used to try to parse strings into literal
        values
    :param parse_keys: Try to parse the keys when they are strings?
    """

    def try_parse(x: Any) -> Any:
        if isinstance(x, str):
            try:
                return parse(x)
            except:  # noqa
                pass
        return x

    if parse_keys:
        def parse_tuple(key: str, value: str) -> Tuple[Any, Any]:
            return try_parse(key), try_parse(value)
    else:
        def parse_tuple(key: str, value: str) -> Tuple[Any, Any]:
            return key, try_parse(value)

    def parse_pair(pair: Union[str, Tuple[Any, Any]]) -> Tuple[Any, Any]:
        if isinstance(pair, str):
            try:
                k, v = pair.split(sep, 1)
            except ValueError as e:
                raise ValueError(f"{pair} is not like KEY{sep}VALUE") from e
            return parse_tuple(k, v)
        return parse_tuple(*pair)

    try:
        items = items.items()  # type: ignore
    except AttributeError:
        pass

    return dict(map(parse_pair, items))
